@@ -1,5 +1,7 @@
 #include <xml_node.h>
 
+static ID decorate, decorate_bang;
+
 #ifdef DEBUG
 static void debug_node_dealloc(xmlNodePtr x)
 {
@@ -12,36 +14,61 @@ static void debug_node_dealloc(xmlNodePtr x)
 
 static void mark(xmlNodePtr node)
 {
-  rb_gc_mark(DOC_RUBY_OBJECT(node->doc));
+  /* it's OK if the document isn't fully realized (as in XML::Reader). */
+  /* see http://github.com/tenderlove/nokogiri/issues/closed/#issue/95 */
+  if (DOC_RUBY_OBJECT_TEST(node->doc) && DOC_RUBY_OBJECT(node->doc))
+    rb_gc_mark(DOC_RUBY_OBJECT(node->doc));
 }
 
 /* :nodoc: */
-typedef xmlNodePtr (*node_other_func)(xmlNodePtr, xmlNodePtr);
+typedef xmlNodePtr (*pivot_reparentee_func)(xmlNodePtr, xmlNodePtr);
 
 /* :nodoc: */
 static void relink_namespace(xmlNodePtr reparented)
 {
-  // Make sure that our reparented node has the correct namespaces
-  if(reparented->doc != (xmlDocPtr)reparented->parent)
+  xmlNodePtr child;
+
+  /* Avoid segv when relinking against unlinked nodes. */
+  if(!reparented->parent) return;
+
+  /* Make sure that our reparented node has the correct namespaces */
+  if(!reparented->ns && reparented->doc != (xmlDocPtr)reparented->parent)
     xmlSetNs(reparented, reparented->parent->ns);
 
-  // Search our parents for an existing definition
+  /* Search our parents for an existing definition */
   if(reparented->nsDef) {
-    xmlNsPtr ns = xmlSearchNsByHref(
-        reparented->doc,
-        reparented->parent,
-        reparented->nsDef->href
-    );
-    if(ns && ns != reparented->nsDef) reparented->nsDef = NULL;
+    xmlNsPtr curr = reparented->nsDef;
+    xmlNsPtr prev = NULL;
+
+    while(curr) {
+      xmlNsPtr ns = xmlSearchNsByHref(
+          reparented->doc,
+          reparented->parent,
+          curr->href
+      );
+      /* If we find the namespace is already declared, remove it from this
+       * definition list. */
+      if(ns && ns != curr) {
+        if (prev) {
+          prev->next = curr->next;
+        } else {
+          reparented->nsDef = curr->next;
+        }
+        NOKOGIRI_ROOT_NSDEF(curr, reparented->doc);
+      } else {
+        prev = curr;
+      }
+      curr = curr->next;
+    }
   }
 
-  // Only walk all children if there actually is a namespace we need to
-  // reparent.
+  /* Only walk all children if there actually is a namespace we need to */
+  /* reparent. */
   if(NULL == reparented->ns) return;
 
-  // When a node gets reparented, walk it's children to make sure that
-  // their namespaces are reparented as well.
-  xmlNodePtr child = reparented->children;
+  /* When a node gets reparented, walk it's children to make sure that */
+  /* their namespaces are reparented as well. */
+  child = reparented->children;
   while(NULL != child) {
     relink_namespace(child);
     child = child->next;
@@ -49,54 +76,121 @@ static void relink_namespace(xmlNodePtr reparented)
 }
 
 /* :nodoc: */
-static VALUE reparent_node_with(VALUE node_obj, VALUE other_obj, node_other_func func)
+static xmlNodePtr xmlReplaceNodeWrapper(xmlNodePtr pivot, xmlNodePtr new_node)
+{
+  xmlNodePtr retval ;
+
+  retval = xmlReplaceNode(pivot, new_node) ;
+
+  if (retval == pivot) {
+    retval = new_node ; /* return semantics for reparent_node_with */
+  }
+
+  /* work around libxml2 issue: https://bugzilla.gnome.org/show_bug.cgi?id=615612 */
+  if (retval->type == XML_TEXT_NODE) {
+    if (retval->prev && retval->prev->type == XML_TEXT_NODE) {
+      retval = xmlTextMerge(retval->prev, retval);
+    }
+    if (retval->next && retval->next->type == XML_TEXT_NODE) {
+      retval = xmlTextMerge(retval, retval->next);
+    }
+  }
+
+  return retval ;
+}
+
+/* :nodoc: */
+static VALUE reparent_node_with(VALUE pivot_obj, VALUE reparentee_obj, pivot_reparentee_func prf)
 {
   VALUE reparented_obj ;
-  xmlNodePtr node, other, reparented ;
+  xmlNodePtr reparentee, pivot, reparented, next_text, new_next_text ;
 
-  if(! rb_funcall(node_obj, rb_intern("is_a?"), 1, cNokogiriXmlNode))
+  if(!rb_obj_is_kind_of(reparentee_obj, cNokogiriXmlNode))
+    rb_raise(rb_eArgError, "node must be a Nokogiri::XML::Node");
+  if(rb_obj_is_kind_of(reparentee_obj, cNokogiriXmlDocument))
     rb_raise(rb_eArgError, "node must be a Nokogiri::XML::Node");
 
-  Data_Get_Struct(node_obj, xmlNode, node);
-  Data_Get_Struct(other_obj, xmlNode, other);
+  Data_Get_Struct(reparentee_obj, xmlNode, reparentee);
+  Data_Get_Struct(pivot_obj, xmlNode, pivot);
 
-  if (node->doc == other->doc) {
-    xmlUnlinkNode(node) ;
-    if ( node->type == XML_TEXT_NODE
-         && other->type == XML_TEXT_NODE
-         && is_2_6_16() ) {
-      other->content = xmlStrdup(other->content); // we'd rather leak than segfault.
-    }
+  if(XML_DOCUMENT_NODE == reparentee->type || XML_HTML_DOCUMENT_NODE == reparentee->type)
+    rb_raise(rb_eArgError, "cannot reparent a document node");
 
-    if(!(reparented = (*func)(other, node))) {
-      rb_raise(rb_eRuntimeError, "Could not reparent node (1)");
-    }
+  xmlUnlinkNode(reparentee);
 
-  } else {
-    xmlNodePtr duped_node ;
-    // recursively copy to the new document
-    if (!(duped_node = xmlDocCopyNode(node, other->doc, 1))) {
+  if (reparentee->doc != pivot->doc || reparentee->type == XML_TEXT_NODE) {
+    /*
+     *  if the reparentee is a text node, there's a very good chance it will be
+     *  merged with an adjacent text node after being reparented, and in that case
+     *  libxml will free the underlying C struct.
+     *
+     *  since we clearly have a ruby object which references the underlying
+     *  memory, we can't let the C struct get freed. let's pickle the original
+     *  reparentee by rooting it; and then we'll reparent a duplicate of the
+     *  node that we don't care about preserving.
+     *
+     *  alternatively, if the reparentee is from a different document than the
+     *  pivot node, libxml2 is going to get confused about which document's
+     *  "dictionary" the node's strings belong to (this is an otherwise
+     *  uninteresting libxml2 implementation detail). as a result, we cannot
+     *  reparent the actual reparentee, so we reparent a duplicate.
+     */
+    NOKOGIRI_ROOT_NODE(reparentee);
+    if (!(reparentee = xmlDocCopyNode(reparentee, pivot->doc, 1))) {
       rb_raise(rb_eRuntimeError, "Could not reparent node (xmlDocCopyNode)");
     }
-    if(!(reparented = (*func)(other, duped_node))) {
-      rb_raise(rb_eRuntimeError, "Could not reparent node (2)");
-    }
-    xmlUnlinkNode(node);
-    NOKOGIRI_ROOT_NODE(node);
   }
 
-  // the child was a text node that was coalesced. we need to have the object
-  // point at SOMETHING, or we'll totally bomb out.
-  if (reparented != node) {
-    DATA_PTR(node_obj) = reparented ;
+  if (reparentee->type == XML_TEXT_NODE && pivot->next && pivot->next->type == XML_TEXT_NODE) {
+    /*
+     *  libxml merges text nodes in a right-to-left fashion, meaning that if
+     *  there are two text nodes who would be adjacent, the right (or following,
+     *  or next) node will be merged into the left (or preceding, or previous)
+     *  node.
+     *
+     *  and by "merged" I mean the string contents will be concatenated onto the
+     *  left node's contents, and then the node will be freed.
+     *
+     *  which means that if we have a ruby object wrapped around the right node,
+     *  its memory would be freed out from under it.
+     *
+     *  so, we detect this edge case and unlink-and-root the text node before it gets
+     *  merged. then we dup the node and insert that duplicate back into the
+     *  document where the real node was.
+     *
+     *  yes, this is totally lame.
+     */
+    next_text     = pivot->next ;
+    new_next_text = xmlDocCopyNode(next_text, pivot->doc, 1) ;
+
+    xmlUnlinkNode(next_text);
+    NOKOGIRI_ROOT_NODE(next_text);
+
+    xmlAddNextSibling(pivot, new_next_text);
   }
 
-  // Appropriately link in namespaces
+  /* TODO: I really want to remove this.  We shouldn't support 2.6.16 anymore */
+  if ( reparentee->type == XML_TEXT_NODE && pivot->type == XML_TEXT_NODE && is_2_6_16() ) {
+    /* work around a string-handling bug in libxml 2.6.16. we'd rather leak than segfault. */
+    pivot->content = xmlStrdup(pivot->content);
+  }
+
+  if(!(reparented = (*prf)(pivot, reparentee))) {
+    rb_raise(rb_eRuntimeError, "Could not reparent node");
+  }
+
+  /*
+   *  make sure the ruby object is pointed at the just-reparented node, which
+   *  might be a duplicate (see above) or might be the result of merging
+   *  adjacent text nodes.
+   */
+  DATA_PTR(reparentee_obj) = reparented ;
+
   relink_namespace(reparented);
 
   reparented_obj = Nokogiri_wrap_xml_node(Qnil, reparented);
 
-  rb_funcall(reparented_obj, rb_intern("decorate!"), 0);
+  rb_funcall(reparented_obj, decorate_bang, 0);
 
   return reparented_obj ;
 }
@@ -126,7 +220,7 @@ static VALUE pointer_id(VALUE self)
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
 
-  return INT2NUM((int)(node));
+  return INT2NUM((long)(node));
 }
 
 /*
@@ -138,18 +232,107 @@ static VALUE pointer_id(VALUE self)
 static VALUE encode_special_chars(VALUE self, VALUE string)
 {
   xmlNodePtr node;
+  xmlChar *encoded;
+  VALUE encoded_str;
+
   Data_Get_Struct(self, xmlNode, node);
-  xmlChar * encoded = xmlEncodeSpecialChars(
+  encoded = xmlEncodeSpecialChars(
       node->doc,
       (const xmlChar *)StringValuePtr(string)
   );
 
-  VALUE encoded_str = NOKOGIRI_STR_NEW2(encoded, node->doc->encoding);
+  encoded_str = NOKOGIRI_STR_NEW2(encoded);
   xmlFree(encoded);
 
   return encoded_str;
 }
 
+/*
+ * call-seq:
+ *  create_internal_subset(name, external_id, system_id)
+ *
+ * Create an internal subset
+ */
+static VALUE create_internal_subset(VALUE self, VALUE name, VALUE external_id, VALUE system_id)
+{
+  xmlNodePtr node;
+  xmlDocPtr doc;
+  xmlDtdPtr dtd;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  doc = node->doc;
+
+  if(xmlGetIntSubset(doc))
+    rb_raise(rb_eRuntimeError, "Document already has an internal subset");
+
+  dtd = xmlCreateIntSubset(
+      doc,
+      NIL_P(name)        ? NULL : (const xmlChar *)StringValuePtr(name),
+      NIL_P(external_id) ? NULL : (const xmlChar *)StringValuePtr(external_id),
+      NIL_P(system_id)   ? NULL : (const xmlChar *)StringValuePtr(system_id)
+  );
+
+  if(!dtd) return Qnil;
+
+  return Nokogiri_wrap_xml_node(Qnil, (xmlNodePtr)dtd);
+}
+
+/*
+ * call-seq:
+ *  create_external_subset(name, external_id, system_id)
+ *
+ * Create an external subset
+ */
+static VALUE create_external_subset(VALUE self, VALUE name, VALUE external_id, VALUE system_id)
+{
+  xmlNodePtr node;
+  xmlDocPtr doc;
+  xmlDtdPtr dtd;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  doc = node->doc;
+
+  if(doc->extSubset)
+    rb_raise(rb_eRuntimeError, "Document already has an external subset");
+
+  dtd = xmlNewDtd(
+      doc,
+      NIL_P(name)        ? NULL : (const xmlChar *)StringValuePtr(name),
+      NIL_P(external_id) ? NULL : (const xmlChar *)StringValuePtr(external_id),
+      NIL_P(system_id)   ? NULL : (const xmlChar *)StringValuePtr(system_id)
+  );
+
+  if(!dtd) return Qnil;
+
+  return Nokogiri_wrap_xml_node(Qnil, (xmlNodePtr)dtd);
+}
+
+/*
+ * call-seq:
+ *  external_subset
+ *
+ * Get the external subset
+ */
+static VALUE external_subset(VALUE self)
+{
+  xmlNodePtr node;
+  xmlDocPtr doc;
+  xmlDtdPtr dtd;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  if(!node->doc) return Qnil;
+
+  doc = node->doc;
+  dtd = doc->extSubset;
+
+  if(!dtd) return Qnil;
+
+  return Nokogiri_wrap_xml_node(Qnil, (xmlNodePtr)dtd);
+}
+
 /*
  * call-seq:
  *  internal_subset
@@ -160,12 +343,14 @@ static VALUE internal_subset(VALUE self)
 {
   xmlNodePtr node;
   xmlDocPtr doc;
+  xmlDtdPtr dtd;
+
   Data_Get_Struct(self, xmlNode, node);
 
   if(!node->doc) return Qnil;
 
   doc = node->doc;
-  xmlDtdPtr dtd = xmlGetIntSubset(doc);
+  dtd = xmlGetIntSubset(doc);
 
   if(!dtd) return Qnil;
 
@@ -182,14 +367,14 @@ static VALUE internal_subset(VALUE self)
 static VALUE duplicate_node(int argc, VALUE *argv, VALUE self)
 {
   VALUE level;
+  xmlNodePtr node, dup;
 
   if(rb_scan_args(argc, argv, "01", &level) == 0)
-    level = INT2NUM(1);
+    level = INT2NUM((long)1);
 
-  xmlNodePtr node, dup;
   Data_Get_Struct(self, xmlNode, node);
 
-  dup = xmlDocCopyNode(node, node->doc, NUM2INT(level));
+  dup = xmlDocCopyNode(node, node->doc, (int)NUM2INT(level));
   if(dup == NULL) return Qnil;
 
   return Nokogiri_wrap_xml_node(rb_obj_class(self), dup);
@@ -220,9 +405,7 @@ static VALUE blank_eh(VALUE self)
 {
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
-  if(1 == xmlIsBlankNode(node))
-    return Qtrue;
-  return Qfalse;
+  return (1 == xmlIsBlankNode(node)) ? Qtrue : Qfalse ;
 }
 
 /*
@@ -259,18 +442,50 @@ static VALUE previous_sibling(VALUE self)
   return Nokogiri_wrap_xml_node(Qnil, sibling);
 }
 
-/* :nodoc: */
-static VALUE replace(VALUE self, VALUE _new_node)
+/*
+ * call-seq:
+ *  next_element
+ *
+ * Returns the next Nokogiri::XML::Element type sibling node.
+ */
+static VALUE next_element(VALUE self)
 {
-  xmlNodePtr node, new_node;
+  xmlNodePtr node, sibling;
   Data_Get_Struct(self, xmlNode, node);
-  Data_Get_Struct(_new_node, xmlNode, new_node);
 
-  xmlReplaceNode(node, new_node);
+  sibling = xmlNextElementSibling(node);
+  if(!sibling) return Qnil;
 
-  // Appropriately link in namespaces
-  relink_namespace(new_node);
-  return self ;
+  return Nokogiri_wrap_xml_node(Qnil, sibling);
+}
+
+/*
+ * call-seq:
+ *  previous_element
+ *
+ * Returns the previous Nokogiri::XML::Element type sibling node.
+ */
+static VALUE previous_element(VALUE self)
+{
+  xmlNodePtr node, sibling;
+  Data_Get_Struct(self, xmlNode, node);
+
+  /*
+   *  note that we don't use xmlPreviousElementSibling here because it's buggy pre-2.7.7.
+   */
+  sibling = node->prev;
+  if(!sibling) return Qnil;
+
+  while(sibling && sibling->type != XML_ELEMENT_NODE)
+    sibling = sibling->prev;
+
+  return sibling ? Nokogiri_wrap_xml_node(Qnil, sibling) : Qnil ;
+}
+
+/* :nodoc: */
+static VALUE replace(VALUE self, VALUE new_node)
+{
+  return reparent_node_with(self, new_node, xmlReplaceNodeWrapper) ;
 }
 
 /*
@@ -282,21 +497,66 @@ static VALUE replace(VALUE self, VALUE _new_node)
 static VALUE children(VALUE self)
 {
   xmlNodePtr node;
+  xmlNodePtr child;
+  xmlNodeSetPtr set;
+  VALUE document;
+  VALUE node_set;
+
   Data_Get_Struct(self, xmlNode, node);
 
-  xmlNodePtr child = node->children;
-  xmlNodeSetPtr set = xmlXPathNodeSetCreate(child);
+  child = node->children;
+  set = xmlXPathNodeSetCreate(child);
 
-  if(!child) return Nokogiri_wrap_xml_node_set(set);
+  document = DOC_RUBY_OBJECT(node->doc);
+
+  if(!child) return Nokogiri_wrap_xml_node_set(set, document);
 
   child = child->next;
   while(NULL != child) {
-    xmlXPathNodeSetAdd(set, child);
+    xmlXPathNodeSetAddUnique(set, child);
     child = child->next;
   }
 
-  VALUE node_set = Nokogiri_wrap_xml_node_set(set);
-  rb_iv_set(node_set, "@document", DOC_RUBY_OBJECT(node->doc));
+  node_set = Nokogiri_wrap_xml_node_set(set, document);
+
+  return node_set;
+}
+
+/*
+ * call-seq:
+ *  element_children
+ *
+ * Get the list of children for this node as a NodeSet.  All nodes will be
+ * element nodes.
+ *
+ * Example:
+ *
+ *   @doc.root.element_children.all? { |x| x.element? } # => true
+ */
+static VALUE element_children(VALUE self)
+{
+  xmlNodePtr node;
+  xmlNodePtr child;
+  xmlNodeSetPtr set;
+  VALUE document;
+  VALUE node_set;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  child = xmlFirstElementChild(node);
+  set = xmlXPathNodeSetCreate(child);
+
+  document = DOC_RUBY_OBJECT(node->doc);
+
+  if(!child) return Nokogiri_wrap_xml_node_set(set, document);
+
+  child = xmlNextElementSibling(child);
+  while(NULL != child) {
+    xmlXPathNodeSetAddUnique(set, child);
+    child = xmlNextElementSibling(child);
+  }
+
+  node_set = Nokogiri_wrap_xml_node_set(set, document);
 
   return node_set;
 }
@@ -318,6 +578,48 @@ static VALUE child(VALUE self)
   return Nokogiri_wrap_xml_node(Qnil, child);
 }
 
+/*
+ * call-seq:
+ *  first_element_child
+ *
+ * Returns the first child node of this node that is an element.
+ *
+ * Example:
+ *
+ *   @doc.root.first_element_child.element? # => true
+ */
+static VALUE first_element_child(VALUE self)
+{
+  xmlNodePtr node, child;
+  Data_Get_Struct(self, xmlNode, node);
+
+  child = xmlFirstElementChild(node);
+  if(!child) return Qnil;
+
+  return Nokogiri_wrap_xml_node(Qnil, child);
+}
+
+/*
+ * call-seq:
+ *  last_element_child
+ *
+ * Returns the last child node of this node that is an element.
+ *
+ * Example:
+ *
+ *   @doc.root.last_element_child.element? # => true
+ */
+static VALUE last_element_child(VALUE self)
+{
+  xmlNodePtr node, child;
+  Data_Get_Struct(self, xmlNode, node);
+
+  child = xmlLastElementChild(node);
+  if(!child) return Qnil;
+
+  return Nokogiri_wrap_xml_node(Qnil, child);
+}
+
 /*
  * call-seq:
  *  key?(attribute)
@@ -344,7 +646,7 @@ static VALUE namespaced_key_eh(VALUE self, VALUE attribute, VALUE namespace)
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
   if(xmlHasNsProp(node, (xmlChar *)StringValuePtr(attribute),
-        Qnil == namespace ? NULL : (xmlChar *)StringValuePtr(namespace)))
+        NIL_P(namespace) ? NULL : (xmlChar *)StringValuePtr(namespace)))
     return Qtrue;
   return Qfalse;
 }
@@ -379,13 +681,13 @@ static VALUE get(VALUE self, VALUE attribute)
   VALUE rval ;
   Data_Get_Struct(self, xmlNode, node);
 
-  if(attribute == Qnil) return Qnil;
+  if(NIL_P(attribute)) return Qnil;
 
   propstr = xmlGetProp(node, (xmlChar *)StringValuePtr(attribute));
 
-  if(NULL == propstr) return Qnil;
+  if(!propstr) return Qnil;
 
-  rval = NOKOGIRI_STR_NEW2(propstr, node->doc->encoding);
+  rval = NOKOGIRI_STR_NEW2(propstr);
 
   xmlFree(propstr);
   return rval ;
@@ -400,10 +702,12 @@ static VALUE get(VALUE self, VALUE attribute)
 static VALUE set_namespace(VALUE self, VALUE namespace)
 {
   xmlNodePtr node;
-  xmlNsPtr ns;
+  xmlNsPtr ns = NULL;
 
   Data_Get_Struct(self, xmlNode, node);
-  Data_Get_Struct(namespace, xmlNs, ns);
+
+  if(!NIL_P(namespace))
+    Data_Get_Struct(namespace, xmlNs, ns);
 
   xmlSetNs(node, ns);
 
@@ -439,7 +743,7 @@ static VALUE attribute_with_ns(VALUE self, VALUE name, VALUE namespace)
   xmlAttrPtr prop;
   Data_Get_Struct(self, xmlNode, node);
   prop = xmlHasNsProp(node, (xmlChar *)StringValuePtr(name),
-      Qnil == namespace ? NULL : (xmlChar *)StringValuePtr(namespace));
+      NIL_P(namespace) ? NULL : (xmlChar *)StringValuePtr(namespace));
 
   if(! prop) return Qnil;
   return Nokogiri_wrap_xml_node(Qnil, (xmlNodePtr)prop);
@@ -454,12 +758,12 @@ static VALUE attribute_with_ns(VALUE self, VALUE name, VALUE namespace)
 static VALUE attribute_nodes(VALUE self)
 {
     /* this code in the mode of xmlHasProp() */
-    xmlNodePtr node ;
-    VALUE attr ;
+    xmlNodePtr node;
+    VALUE attr;
 
-    attr = rb_ary_new() ;
     Data_Get_Struct(self, xmlNode, node);
 
+    attr = rb_ary_new();
     Nokogiri_xml_node_properties(node, attr);
 
     return attr ;
@@ -493,12 +797,14 @@ static VALUE namespace_definitions(VALUE self)
 {
   /* this code in the mode of xmlHasProp() */
   xmlNodePtr node ;
+  VALUE list;
+  xmlNsPtr ns;
 
   Data_Get_Struct(self, xmlNode, node);
 
-  VALUE list = rb_ary_new();
+  list = rb_ary_new();
 
-  xmlNsPtr ns = node->nsDef;
+  ns = node->nsDef;
 
   if(!ns) return list;
 
@@ -510,6 +816,35 @@ static VALUE namespace_definitions(VALUE self)
   return list;
 }
 
+/*
+ *  call-seq:
+ *    namespace_scopes()
+ *
+ *  returns a list of Namespace nodes in scope for _self_. this is all
+ *  namespaces defined in the node, or in any ancestor node.
+ */
+static VALUE namespace_scopes(VALUE self)
+{
+  xmlNodePtr node ;
+  VALUE list;
+  xmlNsPtr *ns_list;
+  int j;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  list = rb_ary_new();
+  ns_list = xmlGetNsList(node->doc, node);
+
+  if(!ns_list) return list;
+
+  for (j = 0 ; ns_list[j] != NULL ; ++j) {
+    rb_ary_push(list, Nokogiri_wrap_xml_namespace(node->doc, ns_list[j]));
+  }
+
+  xmlFree(ns_list);
+  return list;
+}
+
 /*
  * call-seq:
  *  node_type
@@ -520,7 +855,7 @@ static VALUE node_type(VALUE self)
 {
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
-  return INT2NUM((int)node->type);
+  return INT2NUM((long)node->type);
 }
 
 /*
@@ -531,8 +866,17 @@ static VALUE node_type(VALUE self)
  */
 static VALUE set_content(VALUE self, VALUE content)
 {
-  xmlNodePtr node;
+  xmlNodePtr node, child, next ;
   Data_Get_Struct(self, xmlNode, node);
+
+  child = node->children;
+  while (NULL != child) {
+    next = child->next ;
+    xmlUnlinkNode(child) ;
+    NOKOGIRI_ROOT_NODE(child) ;
+    child = next ;
+  }
+
   xmlNodeSetContent(node, (xmlChar *)StringValuePtr(content));
   return content;
 }
@@ -546,26 +890,23 @@ static VALUE set_content(VALUE self, VALUE content)
 static VALUE get_content(VALUE self)
 {
   xmlNodePtr node;
+  xmlChar * content;
+
   Data_Get_Struct(self, xmlNode, node);
 
-  xmlChar * content = xmlNodeGetContent(node);
+  content = xmlNodeGetContent(node);
   if(content) {
-    VALUE rval = NOKOGIRI_STR_NEW2(content, node->doc->encoding);
+    VALUE rval = NOKOGIRI_STR_NEW2(content);
     xmlFree(content);
     return rval;
   }
   return Qnil;
 }
 
-/*
- * call-seq:
- *  add_child(node)
- *
- * Add +node+ as a child of this node. Returns the new child node.
- */
-static VALUE add_child(VALUE self, VALUE child)
+/* :nodoc: */
+static VALUE add_child(VALUE self, VALUE new_child)
 {
-  return reparent_node_with(child, self, xmlAddChild);
+  return reparent_node_with(self, new_child, xmlAddChild);
 }
 
 /*
@@ -610,7 +951,7 @@ static VALUE get_name(VALUE self)
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
   if(node->name)
-    return NOKOGIRI_STR_NEW2(node->name, node->doc->encoding);
+    return NOKOGIRI_STR_NEW2(node->name);
   return Qnil;
 }
 
@@ -624,34 +965,26 @@ static VALUE path(VALUE self)
 {
   xmlNodePtr node;
   xmlChar *path ;
+  VALUE rval;
+
   Data_Get_Struct(self, xmlNode, node);
-  
+
   path = xmlGetNodePath(node);
-  VALUE rval = NOKOGIRI_STR_NEW2(path, node->doc->encoding);
+  rval = NOKOGIRI_STR_NEW2(path);
   xmlFree(path);
   return rval ;
 }
 
-/*
- *  call-seq:
- *    add_next_sibling(node)
- *
- *  Insert +node+ after this node (as a sibling).
- */
-static VALUE add_next_sibling(VALUE self, VALUE rb_node)
+/* :nodoc: */
+static VALUE add_next_sibling(VALUE self, VALUE new_sibling)
 {
-  return reparent_node_with(rb_node, self, xmlAddNextSibling) ;
+  return reparent_node_with(self, new_sibling, xmlAddNextSibling) ;
 }
 
-/*
- * call-seq:
- *  add_previous_sibling(node)
- *
- * Insert +node+ before this node (as a sibling).
- */
-static VALUE add_previous_sibling(VALUE self, VALUE rb_node)
+/* :nodoc: */
+static VALUE add_previous_sibling(VALUE self, VALUE new_sibling)
 {
-  return reparent_node_with(rb_node, self, xmlAddPrevSibling) ;
+  return reparent_node_with(self, new_sibling, xmlAddPrevSibling) ;
 }
 
 /*
@@ -668,23 +1001,29 @@ static VALUE native_write_to(
     VALUE options
 ) {
   xmlNodePtr node;
+  const char * before_indent;
+  xmlSaveCtxtPtr savectx;
 
   Data_Get_Struct(self, xmlNode, node);
 
   xmlIndentTreeOutput = 1;
 
+  before_indent = xmlTreeIndentString;
+
   xmlTreeIndentString = StringValuePtr(indent_string);
 
-  xmlSaveCtxtPtr savectx = xmlSaveToIO(
+  savectx = xmlSaveToIO(
       (xmlOutputWriteCallback)io_write_callback,
       (xmlOutputCloseCallback)io_close_callback,
       (void *)io,
       RTEST(encoding) ? StringValuePtr(encoding) : NULL,
-      NUM2INT(options)
+      (int)NUM2INT(options)
   );
 
   xmlSaveTree(savectx, node);
   xmlSaveClose(savectx);
+
+  xmlTreeIndentString = before_indent;
   return io;
 }
 
@@ -699,7 +1038,7 @@ static VALUE line(VALUE self)
   xmlNodePtr node;
   Data_Get_Struct(self, xmlNode, node);
 
-  return INT2NUM(node->line);
+  return INT2NUM(xmlGetLineNo(node));
 }
 
 /*
@@ -710,25 +1049,32 @@ static VALUE line(VALUE self)
  */
 static VALUE add_namespace_definition(VALUE self, VALUE prefix, VALUE href)
 {
-  xmlNodePtr node;
+  xmlNodePtr node, namespacee;
+  xmlNsPtr ns;
+
   Data_Get_Struct(self, xmlNode, node);
+  namespacee = node ;
 
-
-  xmlNsPtr ns = xmlNewNs(
+  ns = xmlSearchNs(
+      node->doc,
       node,
-      (const xmlChar *)StringValuePtr(href),
-      (const xmlChar *)(prefix == Qnil ? NULL : StringValuePtr(prefix))
+      (const xmlChar *)(NIL_P(prefix) ? NULL : StringValuePtr(prefix))
   );
 
   if(!ns) {
-    ns = xmlSearchNs(
-        node->doc,
-        node,
-        (const xmlChar *)(prefix == Qnil ? NULL : StringValuePtr(prefix))
+    if (node->type != XML_ELEMENT_NODE) {
+      namespacee = node->parent;
+    }
+    ns = xmlNewNs(
+        namespacee,
+        (const xmlChar *)StringValuePtr(href),
+        (const xmlChar *)(NIL_P(prefix) ? NULL : StringValuePtr(prefix))
     );
   }
 
-  if(Qnil == prefix) xmlSetNs(node, ns);
+  if (!ns) return Qnil ;
+
+  if(NIL_P(prefix) || node != namespacee) xmlSetNs(node, ns);
 
   return Nokogiri_wrap_xml_namespace(node->doc, ns);
 }
@@ -742,20 +1088,25 @@ static VALUE add_namespace_definition(VALUE self, VALUE prefix, VALUE href)
 static VALUE new(int argc, VALUE *argv, VALUE klass)
 {
   xmlDocPtr doc;
+  xmlNodePtr node;
   VALUE name;
   VALUE document;
   VALUE rest;
+  VALUE rb_node;
 
   rb_scan_args(argc, argv, "2*", &name, &document, &rest);
 
   Data_Get_Struct(document, xmlDoc, doc);
 
-  xmlNodePtr node = xmlNewNode(NULL, (xmlChar *)StringValuePtr(name));
+  node = xmlNewNode(NULL, (xmlChar *)StringValuePtr(name));
   node->doc = doc->doc;
   NOKOGIRI_ROOT_NODE(node);
 
-  VALUE rb_node = Nokogiri_wrap_xml_node(klass, node);
-  rb_funcall2(rb_node, rb_intern("initialize"), argc, argv);
+  rb_node = Nokogiri_wrap_xml_node(
+      klass == cNokogiriXmlNode ? (VALUE)NULL : klass,
+      node
+  );
+  rb_obj_call_init(rb_node, argc, argv);
 
   if(rb_block_given_p()) rb_yield(rb_node);
 
@@ -772,14 +1123,13 @@ static VALUE dump_html(VALUE self)
 {
   xmlBufferPtr buf ;
   xmlNodePtr node ;
-  Data_Get_Struct(self, xmlNode, node);
+  VALUE html;
 
-  if(node->doc->type == XML_DOCUMENT_NODE)
-    return rb_funcall(self, rb_intern("to_xml"), 0);
+  Data_Get_Struct(self, xmlNode, node);
 
   buf = xmlBufferCreate() ;
   htmlNodeDump(buf, node->doc, node);
-  VALUE html = NOKOGIRI_STR_NEW2(buf->content, node->doc->encoding);
+  html = NOKOGIRI_STR_NEW2(buf->content);
   xmlBufferFree(buf);
   return html ;
 }
@@ -796,17 +1146,79 @@ static VALUE compare(VALUE self, VALUE _other)
   Data_Get_Struct(self, xmlNode, node);
   Data_Get_Struct(_other, xmlNode, other);
 
-  return INT2NUM(xmlXPathCmpNodes(other, node));
+  return INT2NUM((long)xmlXPathCmpNodes(other, node));
 }
 
+
+/* TODO: DOCUMENT ME */
+static VALUE in_context(VALUE self, VALUE _str, VALUE _options)
+{
+  xmlNodePtr node;
+  xmlNodePtr list;
+  xmlNodeSetPtr set;
+  xmlParserErrors error;
+  VALUE doc, err;
+
+  Data_Get_Struct(self, xmlNode, node);
+
+  doc = DOC_RUBY_OBJECT(node->doc);
+  err = rb_iv_get(doc, "@errors");
+
+  xmlSetStructuredErrorFunc((void *)err, Nokogiri_error_array_pusher);
+
+  /* Twiddle global variable because of a bug in libxml2.
+   * http://git.gnome.org/browse/libxml2/commit/?id=e20fb5a72c83cbfc8e4a8aa3943c6be8febadab7
+   */
+#ifndef HTML_PARSE_NOIMPLIED
+  htmlHandleOmittedElem(0);
+#endif
+
+  error = xmlParseInNodeContext(
+      node,
+      StringValuePtr(_str),
+      (int)RSTRING_LEN(_str),
+      (int)NUM2INT(_options),
+      &list);
+
+#ifndef HTML_PARSE_NOIMPLIED
+  htmlHandleOmittedElem(1);
+#endif
+
+  xmlSetStructuredErrorFunc(NULL, NULL);
+
+  /* FIXME: This probably needs to handle more constants... */
+  switch(error) {
+    case XML_ERR_OK:
+      break;
+
+    case XML_ERR_INTERNAL_ERROR:
+    case XML_ERR_NO_MEMORY:
+      rb_raise(rb_eRuntimeError, "error parsing fragment (%d)", error);
+      break;
+
+    default:
+      break;
+  }
+
+  set = xmlXPathNodeSetCreate(NULL);
+
+  while(list) {
+    xmlXPathNodeSetAddUnique(set, list);
+    list = list->next;
+  }
+
+  return Nokogiri_wrap_xml_node_set(set, doc);
+}
+
+
 VALUE Nokogiri_wrap_xml_node(VALUE klass, xmlNodePtr node)
 {
-  assert(node);
-
   VALUE document = Qnil ;
   VALUE node_cache = Qnil ;
   VALUE rb_node = Qnil ;
 
+  assert(node);
+
   if(node->type == XML_DOCUMENT_NODE || node->type == XML_HTML_DOCUMENT_NODE)
       return DOC_RUBY_OBJECT(node->doc);
 
@@ -839,13 +1251,19 @@ VALUE Nokogiri_wrap_xml_node(VALUE klass, xmlNodePtr node)
       klass = cNokogiriXmlProcessingInstruction;
       break;
     case XML_ENTITY_DECL:
-      klass = cNokogiriXmlEntityDeclaration;
+      klass = cNokogiriXmlEntityDecl;
       break;
     case XML_CDATA_SECTION_NODE:
       klass = cNokogiriXmlCData;
       break;
     case XML_DTD_NODE:
-      klass = rb_const_get(mNokogiriXml, rb_intern("DTD"));
+      klass = cNokogiriXmlDtd;
+      break;
+    case XML_ATTRIBUTE_DECL:
+      klass = cNokogiriXmlAttributeDecl;
+      break;
+    case XML_ELEMENT_DECL:
+      klass = cNokogiriXmlElementDecl;
       break;
     default:
       klass = cNokogiriXmlNode;
@@ -856,13 +1274,14 @@ VALUE Nokogiri_wrap_xml_node(VALUE klass, xmlNodePtr node)
   node->_private = (void *)rb_node;
 
   if (DOC_RUBY_OBJECT_TEST(node->doc) && DOC_RUBY_OBJECT(node->doc)) {
+    /* it's OK if the document isn't fully realized (as in XML::Reader). */
+    /* see http://github.com/tenderlove/nokogiri/issues/closed/#issue/95 */
     document = DOC_RUBY_OBJECT(node->doc);
     node_cache = DOC_NODE_CACHE(node->doc);
+    rb_ary_push(node_cache, rb_node);
+    rb_funcall(document, decorate, 1, rb_node);
   }
 
-  rb_ary_push(node_cache, rb_node);
-  rb_funcall(document, rb_intern("decorate"), 1, rb_node);
-
   return rb_node ;
 }
 
@@ -879,7 +1298,6 @@ void Nokogiri_xml_node_properties(xmlNodePtr node, VALUE attr_list)
 
 VALUE cNokogiriXmlNode ;
 VALUE cNokogiriXmlElement ;
-VALUE cNokogiriXmlEntityDeclaration ;
 
 void init_xml_node()
 {
@@ -890,8 +1308,6 @@ void init_xml_node()
   cNokogiriXmlNode = klass;
 
   cNokogiriXmlElement = rb_define_class_under(xml, "Element", klass);
-  cNokogiriXmlEntityDeclaration =
-    rb_define_class_under(xml, "EntityDeclaration", klass);
 
   rb_define_singleton_method(klass, "new", new, -1);
 
@@ -899,12 +1315,16 @@ void init_xml_node()
   rb_define_method(klass, "node_name", get_name, 0);
   rb_define_method(klass, "document", document, 0);
   rb_define_method(klass, "node_name=", set_name, 1);
-  rb_define_method(klass, "add_child", add_child, 1);
   rb_define_method(klass, "parent", get_parent, 0);
   rb_define_method(klass, "child", child, 0);
+  rb_define_method(klass, "first_element_child", first_element_child, 0);
+  rb_define_method(klass, "last_element_child", last_element_child, 0);
   rb_define_method(klass, "children", children, 0);
+  rb_define_method(klass, "element_children", element_children, 0);
   rb_define_method(klass, "next_sibling", next_sibling, 0);
   rb_define_method(klass, "previous_sibling", previous_sibling, 0);
+  rb_define_method(klass, "next_element", next_element, 0);
+  rb_define_method(klass, "previous_element", previous_element, 0);
   rb_define_method(klass, "node_type", node_type, 0);
   rb_define_method(klass, "content", get_content, 0);
   rb_define_method(klass, "path", path, 0);
@@ -917,20 +1337,29 @@ void init_xml_node()
   rb_define_method(klass, "attribute_with_ns", attribute_with_ns, 2);
   rb_define_method(klass, "namespace", namespace, 0);
   rb_define_method(klass, "namespace_definitions", namespace_definitions, 0);
-  rb_define_method(klass, "add_previous_sibling", add_previous_sibling, 1);
-  rb_define_method(klass, "add_next_sibling", add_next_sibling, 1);
+  rb_define_method(klass, "namespace_scopes", namespace_scopes, 0);
   rb_define_method(klass, "encode_special_chars", encode_special_chars, 1);
   rb_define_method(klass, "dup", duplicate_node, -1);
   rb_define_method(klass, "unlink", unlink_node, 0);
   rb_define_method(klass, "internal_subset", internal_subset, 0);
+  rb_define_method(klass, "external_subset", external_subset, 0);
+  rb_define_method(klass, "create_internal_subset", create_internal_subset, 3);
+  rb_define_method(klass, "create_external_subset", create_external_subset, 3);
   rb_define_method(klass, "pointer_id", pointer_id, 0);
   rb_define_method(klass, "line", line, 0);
 
+  rb_define_private_method(klass, "in_context", in_context, 2);
+  rb_define_private_method(klass, "add_child_node", add_child, 1);
+  rb_define_private_method(klass, "add_previous_sibling_node", add_previous_sibling, 1);
+  rb_define_private_method(klass, "add_next_sibling_node", add_next_sibling, 1);
+  rb_define_private_method(klass, "replace_node", replace, 1);
   rb_define_private_method(klass, "dump_html", dump_html, 0);
   rb_define_private_method(klass, "native_write_to", native_write_to, 4);
-  rb_define_private_method(klass, "replace_with_node", replace, 1);
   rb_define_private_method(klass, "native_content=", set_content, 1);
   rb_define_private_method(klass, "get", get, 1);
   rb_define_private_method(klass, "set_namespace", set_namespace, 1);
   rb_define_private_method(klass, "compare", compare, 1);
+
+  decorate      = rb_intern("decorate");
+  decorate_bang = rb_intern("decorate!");
 }
