@@ -1,6 +1,7 @@
 require 'nokogiri/html/entity_lookup'
 require 'nokogiri/html/document'
 require 'nokogiri/html/document_fragment'
+require 'nokogiri/html/sax/parser_context'
 require 'nokogiri/html/sax/parser'
 require 'nokogiri/html/element_description'
 
@@ -23,8 +24,8 @@ module Nokogiri
 
       ####
       # Parse a fragment from +string+ in to a NodeSet.
-      def fragment string
-        HTML::DocumentFragment.parse(string)
+      def fragment string, encoding = nil
+        HTML::DocumentFragment.parse string, encoding
       end
     end
 
