@@ -34,6 +34,9 @@ module Nokogiri
     #
     # You may search this node's subtree using Node#xpath and Node#css
     class Node
+      include Nokogiri::XML::PP::Node
+      include Enumerable
+
       # Element node type, see Nokogiri::XML::Node#element?
       ELEMENT_NODE =       1
       # Attribute node type
@@ -77,7 +80,7 @@ module Nokogiri
       # DOCB document node type
       DOCB_DOCUMENT_NODE = 21
 
-      def initialize name, document
+      def initialize name, document # :nodoc:
         # ... Ya.  This is empty on purpose.
       end
 
@@ -137,15 +140,11 @@ module Nokogiri
         sets = paths.map { |path|
           ctx = XPathContext.new(self)
           ctx.register_namespaces(ns)
-          set = ctx.evaluate(path, handler).node_set
-          set.document = document
-          document.decorate(set)
-          set
+          ctx.evaluate(path, handler)
         }
         return sets.first if sets.length == 1
 
         NodeSet.new(document) do |combined|
-          document.decorate(combined)
           sets.each do |set|
             set.each do |node|
               combined << node
@@ -190,14 +189,40 @@ module Nokogiri
         xpath(*rules)
       end
 
+      ###
+      # Search this node's immediate children using CSS selector +selector+
+      def > selector
+        ns = document.root.namespaces
+        xpath CSS.xpath_for(selector, :prefix => "./", :ns => ns).first
+      end
+
       ###
       # Search for the first occurrence of +path+.
+      #
       # Returns nil if nothing is found, otherwise a Node.
       def at path, ns = document.root ? document.root.namespaces : {}
         search(path, ns).first
       end
       alias :% :at
 
+      ##
+      # Search this node for the first occurrence of XPath +paths+.
+      # Equivalent to <tt>xpath(paths).first</tt>
+      # See Node#xpath for more information.
+      #
+      def at_xpath *paths
+        xpath(*paths).first
+      end
+
+      ##
+      # Search this node for the first occurrence of CSS +rules+.
+      # Equivalent to <tt>css(rules).first</tt>
+      # See Node#css for more information.
+      #
+      def at_css *rules
+        css(*rules).first
+      end
+
       ###
       # Get the attribute value for the attribute +name+
       def [] name
@@ -205,10 +230,137 @@ module Nokogiri
         get(name.to_s)
       end
 
+      ###
+      # Add +node_or_tags+ as a child of this Node.
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the new child node.
+      def add_child node_or_tags
+        node_or_tags = coerce(node_or_tags)
+        if node_or_tags.is_a?(XML::NodeSet)
+          node_or_tags.each { |n| add_child_node n }
+        else
+          add_child_node node_or_tags
+        end
+      end
+
+      ###
+      # Insert +node_or_tags+ before this Node (as a sibling).
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the new sibling node.
+      #
+      # Also see related method +before+.
+      def add_previous_sibling node_or_tags
+        node_or_tags = coerce(node_or_tags)
+        if node_or_tags.is_a?(XML::NodeSet)
+          node_or_tags.each { |n| add_previous_sibling_node n }
+        else
+          add_previous_sibling_node node_or_tags
+        end
+      end
+
+      ###
+      # Insert +node_or_tags+ after this Node (as a sibling).
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the new sibling node.
+      #
+      # Also see related method +after+.
+      def add_next_sibling node_or_tags
+        node_or_tags = coerce(node_or_tags)
+        if node_or_tags.is_a?(XML::NodeSet)
+          if '1.8.6' == RUBY_VERSION
+            node_or_tags.reverse.each { |n| add_next_sibling_node n }
+          else
+            node_or_tags.reverse_each { |n| add_next_sibling_node n }
+          end
+        else
+          add_next_sibling_node node_or_tags
+        end
+      end
+
+      ####
+      # Insert +node_or_tags+ before this node (as a sibling).
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns self, to support chaining of calls.
+      #
+      # Also see related method +add_previous_sibling+.
+      def before node_or_tags
+        add_previous_sibling node_or_tags
+        self
+      end
+
+      ####
+      # Insert +node_or_tags+ after this node (as a sibling).
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a Nokogiri::XML::DocumentFragment, or a string containing markup.
+      #
+      # Returns self, to support chaining of calls.
+      #
+      # Also see related method +add_next_sibling+.
+      def after node_or_tags
+        add_next_sibling node_or_tags
+        self
+      end
+
+      ####
+      # Set the inner_html for this Node to +node_or_tags+
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a Nokogiri::XML::DocumentFragment, or a string containing markup.
+      #
+      # Returns self.
+      def inner_html= node_or_tags
+        node_or_tags = coerce(node_or_tags)
+        children.unlink
+        if node_or_tags.is_a?(XML::NodeSet)
+          node_or_tags.each { |n| add_child_node n }
+        else
+          add_child node_or_tags
+        end
+        self
+      end
+
+      ####
+      # Replace this Node with +node_or_tags+.
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the new child node.
+      #
+      # Also see related method +swap+.
+      def replace node_or_tags
+        node_or_tags = coerce(node_or_tags)
+        if node_or_tags.is_a?(XML::NodeSet)
+          node_or_tags.each { |n| add_previous_sibling n }
+          unlink
+        else
+          replace_node node_or_tags
+        end
+      end
+
+      ####
+      # Swap this Node for +node_or_tags+
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns self, to support chaining of calls.
+      #
+      # Also see related method +replace+.
+      def swap node_or_tags
+        replace node_or_tags
+        self
+      end
+
       alias :next           :next_sibling
       alias :previous       :previous_sibling
+
+      # :stopdoc:
+      # HACK: This is to work around an RDoc bug
+      alias :next=          :add_next_sibling
+      # :startdoc:
+
+      alias :previous=      :add_previous_sibling
       alias :remove         :unlink
       alias :get_attribute  :[]
+      alias :attr           :[]
       alias :set_attribute  :[]=
       alias :text           :content
       alias :inner_text     :content
@@ -219,10 +371,14 @@ module Nokogiri
       alias :type           :node_type
       alias :to_str         :text
       alias :clone          :dup
+      alias :elements       :element_children
 
       ####
-      # Returns a hash containing the node's attributes.  The key is the
-      # attribute name, the value is the string value of the attribute.
+      # Returns a hash containing the node's attributes.  The key is
+      # the attribute name without any namespace, the value is a Nokogiri::XML::Attr
+      # representing the attribute.
+      # If you need to distinguish attributes with the same name, with different namespaces
+      # use #attribute_nodes instead.
       def attributes
         Hash[*(attribute_nodes.map { |node|
           [node.node_name, node]
@@ -245,7 +401,7 @@ module Nokogiri
       # Iterate over each attribute name and value pair for this Node.
       def each &block
         attribute_nodes.each { |node|
-          block.call(node.node_name, node.value)
+          block.call([node.node_name, node.value])
         }
       end
 
@@ -259,55 +415,38 @@ module Nokogiri
       ###
       # Returns true if this Node matches +selector+
       def matches? selector
-        document.search(selector).include?(self)
+        ancestors.last.search(selector).include?(self)
       end
 
-      ####
-      # Create nodes from +data+ and insert them before this node
-      # (as a sibling).
-      def before data
-        fragment(data).children.each do |node|
-          add_previous_sibling node
+      ###
+      # Create a DocumentFragment containing +tags+ that is relative to _this_
+      # context node.
+      def fragment tags
+        type = document.html? ? Nokogiri::HTML : Nokogiri::XML
+        type::DocumentFragment.new(document, tags, self)
+      end
+
+      ###
+      # Parse +string_or_io+ as a document fragment within the context of
+      # *this* node.  Returns a XML::NodeSet containing the nodes parsed from
+      # +string_or_io+.
+      def parse string_or_io, options = ParseOptions::DEFAULT_XML
+        if Fixnum === options
+          options = Nokogiri::XML::ParseOptions.new(options)
         end
-        self
+        # Give the options to the user
+        yield options if block_given?
+
+        contents = string_or_io.respond_to?(:read) ?
+          string_or_io.read :
+          string_or_io
+
+        return Nokogiri::XML::NodeSet.new(document) if contents.empty?
+        in_context(contents, options.to_i)
       end
 
       ####
-      # Create nodes from +data+ and insert them after this node
-      # (as a sibling).
-      def after data
-        fragment(data).children.to_a.reverse.each do |node|
-          add_next_sibling node
-        end
-        self
-      end
-
-      ####
-      # Swap this Node for new nodes made from +data+
-      def swap data
-        before(data)
-        remove
-        self
-      end
-
-      ####
-      # Set the inner_html for this Node to +tags+
-      def inner_html= tags
-        children.each { |x| x.remove}
-
-        fragment(tags).children.to_a.reverse.each do |node|
-          add_child node
-        end
-        self
-      end
-
-      def fragment tags # :nodoc:
-        # TODO: deprecate?
-        document.fragment(tags)
-      end
-
-      ####
-      # Set the Node content to +string+.  The content gets XML escaped.
+      # Set the Node's content to a Text node containing +string+. The string gets XML escaped, not interpreted as markup.
       def content= string
         self.native_content = encode_special_chars(string.to_s)
       end
@@ -322,7 +461,7 @@ module Nokogiri
       ###
       # Get a hash containing the Namespace definitions for this Node
       def namespaces
-        Hash[*namespace_definitions.map { |nd|
+        Hash[*namespace_scopes.map { |nd|
           key = ['xmlns', nd.prefix].compact.join(':')
           if RUBY_VERSION >= '1.9' && document.encoding
             begin
@@ -359,6 +498,11 @@ module Nokogiri
         type == TEXT_NODE
       end
 
+      # Returns true if this is a DocumentFragment
+      def fragment?
+        type == DOCUMENT_FRAG_NODE
+      end
+
       ###
       # Fetch the Nokogiri::HTML::ElementDescription for this node.  Returns
       # nil on XML documents and on unknown tags.
@@ -388,8 +532,8 @@ module Nokogiri
       end
 
       # Get the inner_html for this node's Node#children
-      def inner_html
-        children.map { |x| x.to_html }.join
+      def inner_html *args
+        children.map { |x| x.to_html(*args) }.join
       end
 
       # Get the path to this node as a CSS expression
@@ -399,14 +543,6 @@ module Nokogiri
         }.compact.join(' > ')
       end
 
-      #  recursively get all namespaces from this node and its subtree
-      def collect_namespaces
-        # TODO: print warning message if a prefix refers to more than one URI in the document?
-        ns = {}
-        traverse {|j| ns.merge!(j.namespaces)}
-        ns
-      end
-
       ###
       # Get a list of ancestor Node for this Node.  If +selector+ is given,
       # the ancestors must match +selector+
@@ -423,8 +559,10 @@ module Nokogiri
 
         return NodeSet.new(document, parents) unless selector
 
+        root = parents.last
+
         NodeSet.new(document, parents.find_all { |parent|
-          parent.matches?(selector)
+          root.search(selector).include?(parent)
         })
       end
 
@@ -438,12 +576,15 @@ module Nokogiri
       ###
       # Set the namespace for this node to +ns+
       def namespace= ns
+        return set_namespace(ns) unless ns
+
+        unless Nokogiri::XML::Namespace === ns
+          raise TypeError, "#{ns.class} can't be coerced into Nokogiri::XML::Namespace"
+        end
         if ns.document != document
           raise ArgumentError, 'namespace must be declared on the same document'
         end
-        unless ns.is_a? Nokogiri::XML::Namespace
-          raise TypeError, "#{ns.class} can't be coerced into Nokogiri::XML::Namespace"
-        end
+
         set_namespace ns
       end
 
@@ -460,18 +601,6 @@ module Nokogiri
         visitor.visit(self)
       end
 
-      ####
-      #  replace this Node with the +new_node+ in the Document.
-      def replace new_node
-        if new_node.is_a?(Document) || !new_node.is_a?(XML::Node)
-          raise ArgumentError, <<-EOERR
-Node.replace requires a Node argument, and cannot accept a Document.
-(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
-          EOERR
-        end
-        replace_with_node new_node
-      end
-
       ###
       # Test to see if this Node is equal to +other+
       def == other
@@ -481,8 +610,8 @@ Node.replace requires a Node argument, and cannot accept a Document.
       end
 
       ###
-      # Serialize Node using +options+.  Save options 
-      # can also be set using a block. See SaveOptions.
+      # Serialize Node using +options+.  Save options can also be set using a
+      # block. See SaveOptions.
       #
       # These two statements are equivalent:
       #
@@ -500,10 +629,15 @@ Node.replace requires a Node argument, and cannot accept a Document.
           :save_with  => args[1] || SaveOptions::FORMAT
         }
 
-        io = StringIO.new
+        encoding = options[:encoding] || document.encoding
+
+        outstring = ""
+        if encoding && outstring.respond_to?(:force_encoding)
+          outstring.force_encoding(Encoding.find(encoding))
+        end
+        io = StringIO.new(outstring)
         write_to io, options, &block
-        io.rewind
-        io.read
+        io.string
       end
 
       ###
@@ -532,8 +666,6 @@ Node.replace requires a Node argument, and cannot accept a Document.
       #
       # See Node#write_to for a list of +options+
       def to_xml options = {}
-        encoding = nil
-
         options[:save_with] ||= SaveOptions::FORMAT | SaveOptions::AS_XML
 
         serialize(options)
@@ -637,8 +769,27 @@ Node.replace requires a Node argument, and cannot accept a Document.
         return nil unless document == other.document
         compare other
       end
+
+      private
+
+      def coerce data # :nodoc:
+        return data                    if data.is_a?(XML::NodeSet)
+        return data.children           if data.is_a?(XML::DocumentFragment)
+        return fragment(data).children if data.is_a?(String)
+
+        if data.is_a?(Document) || !data.is_a?(XML::Node)
+          raise ArgumentError, <<-EOERR
+Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
+(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
+          EOERR
+        end
+
+        data
+      end
+
+      def inspect_attributes
+        [:name, :namespace, :attribute_nodes, :children]
+      end
     end
   end
 end
-
-class Nokogiri::XML::Element < Nokogiri::XML::Node ; end
