@@ -1,4 +1,5 @@
-require File.expand_path(File.join(File.dirname(__FILE__), '..', '..', "helper"))
+# -*- coding: utf-8 -*-
+require "helper"
 
 module Nokogiri
   module HTML
@@ -9,6 +10,18 @@ module Nokogiri
           @parser = HTML::SAX::Parser.new(Doc.new)
         end
 
+        def test_parse_empty_document
+          # This caused a segfault in libxml 2.6.x
+          assert_nothing_raised { @parser.parse '' }
+        end
+
+        def test_parse_empty_file
+          # Make sure empty files don't break stuff
+          empty_file_name =  File.join(Dir.tmpdir, 'bogus.xml')
+          FileUtils.touch empty_file_name
+          assert_nothing_raised { @parser.parse_file empty_file_name }
+        end
+
         def test_parse_file
           @parser.parse_file(HTML_FILE)
           assert_equal 1110, @parser.document.end_elements.length
@@ -22,7 +35,7 @@ module Nokogiri
 
         def test_parse_file_non_existant
           assert_raise Errno::ENOENT do
-            @parser.parse_file('foo')
+            @parser.parse_file('there_is_no_reasonable_way_this_file_exists')
           end
         end
 
@@ -38,6 +51,15 @@ module Nokogiri
           end
         end
 
+        def test_parse_force_encoding
+          @parser.parse_memory(<<-HTML, 'UTF-8')
+          <meta http-equiv="Content-Type" content="text/html; charset=windows-1251">
+          Информация
+          HTML
+          assert_equal("Информация",
+                       @parser.document.data.join.strip)
+        end
+
         def test_parse_document
           @parser.parse_memory(<<-eoxml)
             <p>Paragraph 1</p>
