@@ -1,4 +1,4 @@
-require File.expand_path(File.join(File.dirname(__FILE__), '..', "helper"))
+require "helper"
 
 require 'stringio'
 
@@ -10,26 +10,91 @@ module Nokogiri
         @xml = Nokogiri::XML(File.read(XML_FILE), XML_FILE)
       end
 
-      def test_namespace_nodes
-        xml = Nokogiri::XML <<-eoxml
-          <root xmlns="http://tenderlovemaking.com/" xmlns:foo="bar">
-            <awesome/>
-          </root>
-        eoxml
-        awesome = xml.root
-        namespaces = awesome.namespace_definitions
-        assert_equal 2, namespaces.length
+      def test_first_element_child
+        node = @xml.root.first_element_child
+        assert_equal 'employee', node.name
+        assert node.element?, 'node is an element'
       end
 
-      def test_no_definitions
-        xml = Nokogiri::XML <<-eoxml
-          <root xmlns="http://tenderlovemaking.com/" xmlns:foo="bar">
-            <awesome/>
-          </root>
-        eoxml
-        awesome = xml.at('//xmlns:awesome')
-        namespaces = awesome.namespace_definitions
-        assert_equal 0, namespaces.length
+      def test_element_children
+        nodes = @xml.root.element_children
+        assert_equal @xml.root.first_element_child, nodes.first
+        assert nodes.all? { |node| node.element? }, 'all nodes are elements'
+      end
+
+      def test_last_element_child
+        nodes = @xml.root.element_children
+        assert_equal nodes.last, @xml.root.element_children.last
+      end
+
+      def test_bad_xpath
+        bad_xpath = '//foo['
+
+        begin
+          @xml.xpath(bad_xpath)
+        rescue Nokogiri::XML::XPath::SyntaxError => e
+          assert_match(bad_xpath, e.to_s)
+        end
+      end
+
+      def test_namespace_type_error
+        assert_raises(TypeError) do
+          @xml.root.namespace = Object.new
+        end
+      end
+
+      def test_remove_namespace
+        @xml = Nokogiri::XML('<r xmlns="v"><s /></r>')
+        tag = @xml.at('s')
+        assert tag.namespace
+        tag.namespace = nil
+        assert_nil tag.namespace
+      end
+
+      def test_parse_needs_doc
+        list = @xml.root.parse('fooooooo <hello />')
+        assert_equal 1, list.css('hello').length
+      end
+
+      def test_parse
+        list = @xml.root.parse('fooooooo <hello />')
+        assert_equal 2, list.length
+      end
+
+      def test_parse_with_block
+        called = false
+        list = @xml.root.parse('<hello />') { |cfg|
+          called = true
+          assert_instance_of Nokogiri::XML::ParseOptions, cfg
+        }
+        assert called, 'config block called'
+        assert_equal 1, list.length
+      end
+
+      def test_parse_with_io
+        list = @xml.root.parse(StringIO.new('<hello />'))
+        assert_equal 1, list.length
+        assert_equal 'hello', list.first.name
+      end
+
+      def test_parse_with_empty_string
+        list = @xml.root.parse('')
+        assert_equal 0, list.length
+      end
+
+      # descriptive, not prescriptive.
+      def test_parse_invalid_html_markup_results_in_empty_nodeset
+        doc = Nokogiri::HTML("<html></html>")
+        nodeset = doc.root.parse "<div><div>a</div><snippet>b</snippet></div>"
+        assert_equal 1, doc.errors.length # "Tag snippet invalid"
+        assert_equal 0, nodeset.length
+      end
+
+      def test_parse_error_list
+        error_count = @xml.errors.length
+        list = @xml.root.parse('<hello>')
+        assert_equal 0, list.length
+        assert(error_count < @xml.errors.length, "errors should have increased")
       end
 
       def test_subclass_dup
@@ -38,6 +103,127 @@ module Nokogiri
         assert_instance_of subclass, node
       end
 
+      def test_gt_string_arg
+        node = @xml.at('employee')
+        nodes = (node > 'name')
+        assert_equal 1, nodes.length
+        assert_equal node, nodes.first.parent
+      end
+
+      def test_next_element_when_next_sibling_is_element_should_return_next_sibling
+        doc = Nokogiri::XML "<root><foo /><quux /></root>"
+        node         = doc.at_css("foo")
+        next_element = node.next_element
+        assert next_element.element?
+        assert_equal doc.at_css("quux"), next_element
+      end
+
+      def test_next_element_when_there_is_no_next_sibling_should_return_nil
+        doc = Nokogiri::XML "<root><foo /><quux /></root>"
+        assert_nil doc.at_css("quux").next_element
+      end
+
+      def test_next_element_when_next_sibling_is_not_an_element_should_return_closest_next_element_sibling
+        doc = Nokogiri::XML "<root><foo />bar<quux /></root>"
+        node         = doc.at_css("foo")
+        next_element = node.next_element
+        assert next_element.element?
+        assert_equal doc.at_css("quux"), next_element
+      end
+
+      def test_next_element_when_next_sibling_is_not_an_element_and_no_following_element_should_return_nil
+        doc = Nokogiri::XML "<root><foo />bar</root>"
+        node         = doc.at_css("foo")
+        next_element = node.next_element
+        assert_nil next_element
+      end
+
+      def test_previous_element_when_previous_sibling_is_element_should_return_previous_sibling
+        doc = Nokogiri::XML "<root><foo /><quux /></root>"
+        node             = doc.at_css("quux")
+        previous_element = node.previous_element
+        assert previous_element.element?
+        assert_equal doc.at_css("foo"), previous_element
+      end
+
+      def test_previous_element_when_there_is_no_previous_sibling_should_return_nil
+        doc = Nokogiri::XML "<root><foo /><quux /></root>"
+        assert_nil doc.at_css("foo").previous_element
+      end
+
+      def test_previous_element_when_previous_sibling_is_not_an_element_should_return_closest_previous_element_sibling
+        doc = Nokogiri::XML "<root><foo />bar<quux /></root>"
+        node             = doc.at_css("quux")
+        previous_element = node.previous_element
+        assert previous_element.element?
+        assert_equal doc.at_css("foo"), previous_element
+      end
+
+      def test_previous_element_when_previous_sibling_is_not_an_element_and_no_following_element_should_return_nil
+        doc = Nokogiri::XML "<root>foo<bar /></root>"
+        node             = doc.at_css("bar")
+        previous_element = node.previous_element
+        assert_nil previous_element
+      end
+
+      def test_element?
+        assert @xml.root.element?, 'is an element'
+      end
+
+      def test_slash_search
+        assert_equal 'EMP0001', (@xml/:staff/:employee/:employeeId).first.text
+      end
+
+      def test_append_with_document
+        assert_raises(ArgumentError) do
+          @xml.root << Nokogiri::XML::Document.new
+        end
+      end
+
+      def test_inspect_ns
+        xml = Nokogiri::XML(<<-eoxml) { |c| c.noblanks }
+          <root xmlns="http://tenderlovemaking.com/" xmlns:foo="bar">
+            <awesome/>
+          </root>
+        eoxml
+        ins = xml.inspect
+
+        xml.traverse do |node|
+          assert_match node.class.name, ins
+          if node.respond_to? :attributes
+            node.attributes.each do |k,v|
+              assert_match k, ins
+              assert_match v, ins
+            end
+          end
+
+          if node.respond_to?(:namespace) && node.namespace
+            assert_match node.namespace.class.name, ins
+            assert_match node.namespace.href, ins
+          end
+        end
+      end
+
+      def test_namespace_definitions_when_some_exist
+        xml = Nokogiri::XML <<-eoxml
+          <root xmlns="http://tenderlovemaking.com/" xmlns:foo="bar">
+            <awesome/>
+          </root>
+        eoxml
+        namespace_definitions = xml.root.namespace_definitions
+        assert_equal 2, namespace_definitions.length
+      end
+
+      def test_namespace_definitions_when_no_exist
+        xml = Nokogiri::XML <<-eoxml
+          <root xmlns="http://tenderlovemaking.com/" xmlns:foo="bar">
+            <awesome/>
+          </root>
+        eoxml
+        namespace_definitions = xml.at_xpath('//xmlns:awesome').namespace_definitions
+        assert_equal 0, namespace_definitions.length
+      end
+
       def test_namespace_goes_to_children
         fruits = Nokogiri::XML(<<-eoxml)
         <Fruit xmlns='www.fruits.org'>
@@ -55,18 +241,6 @@ module Nokogiri
         assert_nil @xml.at('employee').description
       end
 
-      def test_add_namespace_add_child
-        doc   = Nokogiri::XML::Document.new
-        item  = Nokogiri::XML::Element.new('item', doc)
-        doc.root = item
-
-        entry = Nokogiri::XML::Element.new('entry', doc)
-        entry.add_namespace('tlm', 'http://tenderlovemaking.com')
-        assert_equal 'http://tenderlovemaking.com', entry.namespaces['xmlns:tlm']
-        item.add_child(entry)
-        assert_equal 'http://tenderlovemaking.com', entry.namespaces['xmlns:tlm']
-      end
-
       def test_spaceship
         nodes = @xml.xpath('//employee')
         assert_equal(-1, (nodes.first <=> nodes.last))
@@ -114,23 +288,24 @@ module Nokogiri
         end
       end
 
+      def test_fragment_creates_elements
+        apple = @xml.fragment('<Apple/>')
+        apple.children.each do |child|
+          assert_equal Nokogiri::XML::Node::ELEMENT_NODE, child.type
+          assert_instance_of Nokogiri::XML::Element, child
+        end
+      end
+
       def test_node_added_to_root_should_get_namespace
         fruits = Nokogiri::XML(<<-eoxml)
           <Fruit xmlns='http://www.fruits.org'>
           </Fruit>
         eoxml
         apple = fruits.fragment('<Apple/>')
-        fruits << apple
+        fruits.root << apple
         assert_equal 1, fruits.xpath('//xmlns:Apple').length
       end
 
-      def test_add_child_path_following_sequential_text_nodes
-        xml = Nokogiri::XML('<root>text</root>')
-        xml.root.add_child(Nokogiri::XML::Text.new('text', xml))
-        item = xml.root.add_child(Nokogiri::XML::Element.new('item', xml))
-        assert_equal '/root/item', item.path
-      end
-
       def test_new_node_can_have_ancestors
         xml = Nokogiri::XML('<root>text</root>')
         item = Nokogiri::XML::Element.new('item', xml)
@@ -210,7 +385,7 @@ module Nokogiri
       def test_add_namespace_does_not_associate_node
         node = @xml.at('address')
         assert_nil node.namespace
-        definition = node.add_namespace_definition 'foo', 'http://tlm.com/'
+        assert node.add_namespace_definition 'foo', 'http://tlm.com/'
         assert_nil node.namespace
       end
 
@@ -236,6 +411,20 @@ module Nokogiri
         assert_equal node, @xml.xpath('//address').first
       end
 
+      def test_at_xpath
+        node = @xml.at_xpath('//address')
+        nodes = @xml.xpath('//address')
+        assert_equal 5, nodes.size
+        assert_equal node, nodes.first
+      end
+
+      def test_at_css
+        node = @xml.at_css('address')
+        nodes = @xml.css('address')
+        assert_equal 5, nodes.size
+        assert_equal node, nodes.first
+      end
+
       def test_percent
         node = @xml % ('address')
         assert_equal node, @xml.xpath('//address').first
@@ -257,34 +446,6 @@ module Nokogiri
         assert visitor.visited
       end
 
-      def test_replace_with_default_namespaces
-        fruits = Nokogiri::XML(<<-eoxml)
-          <fruit xmlns="http://fruits.org">
-            <apple />
-          </fruit>
-        eoxml
-
-        apple = fruits.css('apple').first
-
-        orange = Nokogiri::XML::Node.new('orange', fruits)
-        apple.replace(orange)
-
-        assert_equal orange, fruits.css('orange').first
-      end
-
-      def test_add_child_should_inherit_namespace
-        doc = Nokogiri::XML(<<-eoxml)
-          <root xmlns="http://tenderlovemaking.com/">
-            <first>
-            </first>
-          </root>
-        eoxml
-        assert node = doc.at('//xmlns:first')
-        child = Nokogiri::XML::Node.new('second', doc)
-        node.add_child(child)
-        assert doc.at('//xmlns:second')
-      end
-
       def test_write_to
         io = StringIO.new
         @xml.write_to io
@@ -364,7 +525,7 @@ module Nokogiri
       def test_new
         assert node = Nokogiri::XML::Node.new('input', @xml)
         assert_equal 1, node.node_type
-        assert_instance_of Nokogiri::XML::Node, node
+        assert_instance_of Nokogiri::XML::Element, node
       end
 
       def test_to_str
@@ -401,110 +562,12 @@ module Nokogiri
         assert_nil address['domestic']
       end
 
-      def test_add_child_in_same_document
-        child = @xml.css('employee').first
-
-        assert previous_last_child = child.children.last
-        assert new_child = child.children.first
-
-        last = child.children.last
-
-        child.add_child(new_child)
-        assert_equal new_child, child.children.last
-        assert_equal last, child.children.last
-      end
-
-      def test_add_child_from_other_document
-        d1 = Nokogiri::XML("<root><item>1</item><item>2</item></root>")
-        d2 = Nokogiri::XML("<root><item>3</item><item>4</item></root>")
-
-        d2.at('root').search('item').each do |i|
-          d1.at('root').add_child i
-        end
-
-        assert_equal 0, d2.search('item').size
-        assert_equal 4, d1.search('item').size
-      end
-
-      def test_add_child
-        xml = Nokogiri::XML(<<-eoxml)
-        <root>
-          <a>Hello world</a>
-        </root>
-        eoxml
-        text_node = Nokogiri::XML::Text.new('hello', xml)
-        assert_equal Nokogiri::XML::Node::TEXT_NODE, text_node.type
-        xml.root.add_child text_node
-        assert_match 'hello', xml.to_s
-      end
-
-      def test_chevron_works_as_add_child
-        xml = Nokogiri::XML(<<-eoxml)
-        <root>
-          <a>Hello world</a>
-        </root>
-        eoxml
-        text_node = Nokogiri::XML::Text.new('hello', xml)
-        xml.root << text_node
-        assert_match 'hello', xml.to_s
-      end
-
       def test_set_content_with_symbol
         node = @xml.at('//name')
         node.content = :foo
         assert_equal 'foo', node.content
       end
 
-      def test_add_previous_sibling
-        xml = Nokogiri::XML(<<-eoxml)
-        <root>
-          <a>Hello world</a>
-        </root>
-        eoxml
-        b_node = Nokogiri::XML::Node.new('a', xml)
-        assert_equal Nokogiri::XML::Node::ELEMENT_NODE, b_node.type
-        b_node.content = 'first'
-        a_node = xml.xpath('//a').first
-        a_node.add_previous_sibling(b_node)
-        assert_equal('first', xml.xpath('//a').first.text)
-      end
-
-      def test_add_previous_sibling_merge
-        xml = Nokogiri::XML(<<-eoxml)
-        <root>
-          <a>Hello world</a>
-        </root>
-        eoxml
-
-        assert a_tag = xml.css('a').first
-
-        left_space = a_tag.previous
-        right_space = a_tag.next
-        assert left_space.text?
-        assert right_space.text?
-
-        left_space.add_previous_sibling(right_space)
-        assert_equal left_space, right_space
-      end
-
-      def test_add_next_sibling_merge
-        xml = Nokogiri::XML(<<-eoxml)
-        <root>
-          <a>Hello world</a>
-        </root>
-        eoxml
-
-        assert a_tag = xml.css('a').first
-
-        left_space = a_tag.previous
-        right_space = a_tag.next
-        assert left_space.text?
-        assert right_space.text?
-
-        right_space.add_next_sibling(left_space)
-        assert_equal left_space, right_space
-      end
-
       def test_find_by_css_with_tilde_eql
         xml = Nokogiri::XML.parse(<<-eoxml)
         <root>
@@ -629,6 +692,17 @@ module Nokogiri
 
         node.content = 'hello world!'
         assert_equal('hello world!', node.content)
+
+        node.content = '& <foo> &amp;'
+        assert_equal('& <foo> &amp;', node.content)
+        assert_equal('<form>&amp; &lt;foo&gt; &amp;amp;</form>', node.to_xml)
+      end
+
+      def test_set_content_should_unlink_existing_content
+        node     = @xml.at_css("employee")
+        children = node.children
+        node.content = "hello"
+        children.each { |child| assert_nil child.parent }
       end
 
       def test_whitespace_nodes
@@ -638,30 +712,6 @@ module Nokogiri
         assert_equal " ", children[3]
       end
 
-      def test_replace
-        set = @xml.search('//employee')
-        assert 5, set.length
-        assert 0, @xml.search('//form').length
-
-        first = set[0]
-        second = set[1]
-
-        node = Nokogiri::XML::Node.new('form', @xml)
-        first.replace(node)
-
-        assert set = @xml.search('//employee')
-        assert_equal 4, set.length
-        assert 1, @xml.search('//form').length
-
-        assert_equal set[0].to_xml, second.to_xml
-      end
-
-      def test_illegal_replace_of_node_with_doc
-        new_node = Nokogiri::XML.parse('<foo>bar</foo>')
-        old_node = @xml.at('//employee')
-        assert_raises(ArgumentError){ old_node.replace new_node }
-      end
-
       def test_node_equality
         doc1 = Nokogiri::XML.parse(File.read(XML_FILE), XML_FILE)
         doc2 = Nokogiri::XML.parse(File.read(XML_FILE), XML_FILE)
@@ -675,16 +725,16 @@ module Nokogiri
         assert_equal address1_1, address1_2 # two references to the exact same node
       end
 
-      def test_namespace_as_hash
+      def test_namespace_search_with_xpath_and_hash
         xml = Nokogiri::XML.parse(<<-eoxml)
-<root>
- <car xmlns:part="http://general-motors.com/">
-  <part:tire>Michelin Model XGV</part:tire>
- </car>
- <bicycle xmlns:part="http://schwinn.com/">
-  <part:tire>I'm a bicycle tire!</part:tire>
- </bicycle>
-</root>
+        <root>
+          <car xmlns:part="http://general-motors.com/">
+            <part:tire>Michelin Model XGV</part:tire>
+          </car>
+          <bicycle xmlns:part="http://schwinn.com/">
+            <part:tire>I'm a bicycle tire!</part:tire>
+          </bicycle>
+        </root>
         eoxml
 
         tires = xml.xpath('//bike:tire', {'bike' => 'http://schwinn.com/'})
@@ -693,61 +743,70 @@ module Nokogiri
 
       def test_namespace_search_with_css
         xml = Nokogiri::XML.parse(<<-eoxml)
-<root>
- <car xmlns:part="http://general-motors.com/">
-  <part:tire>Michelin Model XGV</part:tire>
- </car>
- <bicycle xmlns:part="http://schwinn.com/">
-  <part:tire>I'm a bicycle tire!</part:tire>
- </bicycle>
-</root>
+        <root>
+          <car xmlns:part="http://general-motors.com/">
+            <part:tire>Michelin Model XGV</part:tire>
+          </car>
+          <bicycle xmlns:part="http://schwinn.com/">
+            <part:tire>I'm a bicycle tire!</part:tire>
+          </bicycle>
+        </root>
         eoxml
 
         tires = xml.css('bike|tire', 'bike' => 'http://schwinn.com/')
         assert_equal 1, tires.length
       end
 
-      def test_namespaces
+      def test_namespaces_should_include_all_namespace_definitions
         xml = Nokogiri::XML.parse(<<-EOF)
-<x xmlns:a='http://foo.com/' xmlns:b='http://bar.com/'>
-  <y xmlns:c='http://bazz.com/'>
-    <a:div>hello a</a:div>
-    <b:div>hello b</b:div>
-    <c:div>hello c</c:div>
-  </y>  
-</x>
-EOF
-        assert namespaces = xml.root.namespaces
-        assert namespaces.key?('xmlns:a')
-        assert_equal 'http://foo.com/', namespaces['xmlns:a']
-        assert namespaces.key?('xmlns:b')
-        assert_equal 'http://bar.com/', namespaces['xmlns:b']
-        assert ! namespaces.key?('xmlns:c')
+        <x xmlns="http://quux.com/" xmlns:a="http://foo.com/" xmlns:b="http://bar.com/">
+          <y xmlns:c="http://bazz.com/">
+            <z>hello</z>
+            <a xmlns:c="http://newc.com/" />
+          </y>
+        </x>
+        EOF
 
-        assert namespaces = xml.namespaces
-        assert namespaces.key?('xmlns:a')
-        assert_equal 'http://foo.com/', namespaces['xmlns:a']
-        assert namespaces.key?('xmlns:b')
-        assert_equal 'http://bar.com/', namespaces['xmlns:b']
-        assert namespaces.key?('xmlns:c')
-        assert_equal 'http://bazz.com/', namespaces['xmlns:c']
+        namespaces = xml.namespaces # Document#namespace
+        assert_equal({"xmlns"   => "http://quux.com/",
+                      "xmlns:a" => "http://foo.com/",
+                      "xmlns:b" => "http://bar.com/"}, namespaces)
 
-        assert_equal "hello a", xml.search("//a:div", xml.namespaces).first.inner_text
-        assert_equal "hello b", xml.search("//b:div", xml.namespaces).first.inner_text
-        assert_equal "hello c", xml.search("//c:div", xml.namespaces).first.inner_text
+        namespaces = xml.root.namespaces
+        assert_equal({"xmlns"   => "http://quux.com/",
+                      "xmlns:a" => "http://foo.com/",
+                      "xmlns:b" => "http://bar.com/"}, namespaces)
+
+        namespaces = xml.at_xpath("//xmlns:y").namespaces
+        assert_equal({"xmlns"   => "http://quux.com/",
+                      "xmlns:a" => "http://foo.com/",
+                      "xmlns:b" => "http://bar.com/",
+                      "xmlns:c" => "http://bazz.com/"}, namespaces)
+
+        namespaces = xml.at_xpath("//xmlns:z").namespaces
+        assert_equal({"xmlns"   => "http://quux.com/",
+                      "xmlns:a" => "http://foo.com/",
+                      "xmlns:b" => "http://bar.com/",
+                      "xmlns:c" => "http://bazz.com/"}, namespaces)
+
+        namespaces = xml.at_xpath("//xmlns:a").namespaces
+        assert_equal({"xmlns"   => "http://quux.com/",
+                      "xmlns:a" => "http://foo.com/",
+                      "xmlns:b" => "http://bar.com/",
+                      "xmlns:c" => "http://newc.com/"}, namespaces)
       end
 
       def test_namespace
         xml = Nokogiri::XML.parse(<<-EOF)
-<x xmlns:a='http://foo.com/' xmlns:b='http://bar.com/'>
-  <y xmlns:c='http://bazz.com/'>
-    <a:div>hello a</a:div>
-    <b:div>hello b</b:div>
-    <c:div>hello c</c:div>
-    <div>hello moon</div>
-  </y>  
-</x>
-EOF
+        <x xmlns:a='http://foo.com/' xmlns:b='http://bar.com/'>
+          <y xmlns:c='http://bazz.com/'>
+            <a:div>hello a</a:div>
+            <b:div>hello b</b:div>
+            <c:div>hello c</c:div>
+            <div>hello moon</div>
+          </y>  
+        </x>
+        EOF
         set = xml.search("//y/*")
         assert_equal "a", set[0].namespace.prefix
         assert_equal "b", set[1].namespace.prefix
@@ -758,7 +817,7 @@ EOF
       def test_namespace_without_an_href_on_html_node
         # because microsoft word's HTML formatting does this. ick.
         xml = Nokogiri::HTML.parse <<-EOF
-<div><o:p>foo</o:p></div>
+          <div><o:p>foo</o:p></div>
         EOF
 
         assert_not_nil(node = xml.at('p'))
@@ -782,34 +841,25 @@ EOF
         assert_equal 2, node.line
       end
 
-      def test_unlink_then_reparent
-        # see http://github.com/tenderlove/nokogiri/issues#issue/22
-        10.times do
-          STDOUT.putc "."
-          STDOUT.flush
-          begin
-            doc = Nokogiri::XML <<-EOHTML
-              <root>
-                <a>
-                  <b/>
-                  <c/>
-                </a>
-              </root>
-            EOHTML
-
-            root = doc.at("root")
-            a = root.at("a")
-            b = a.at("b")
-            c = a.at("c")
-            a.add_next_sibling(b.unlink)
-            c.unlink
-          end
-
-          GC.start
+      def test_xpath_results_have_document_and_are_decorated
+        x = Module.new do
+          def awesome! ; end
         end
-
+        util_decorate(@xml, x)
+        node_set = @xml.xpath("//employee")
+        assert_equal @xml, node_set.document
+        assert node_set.respond_to?(:awesome!)
       end
 
+      def test_css_results_have_document_and_are_decorated
+        x = Module.new do
+          def awesome! ; end
+        end
+        util_decorate(@xml, x)
+        node_set = @xml.css("employee")
+        assert_equal @xml, node_set.document
+        assert node_set.respond_to?(:awesome!)
+      end
     end
   end
 end
